/*----------------------------------------------------------------------------
ALi Aladdin V ACPI enabler for Throttle - G. Broers 2014, thanks to Chris Giese.
Throttle C port and integration - Braden "Blzut3" Obrzut 2018
Last code change GB 12-10-2018

Text below from original PCI.C, where this enabler is based on. 

--------------------------------------------------------------------

PCI demo for DJGPP, Turbo C, or Watcom C, using I/O ports
(configuration mechanism 1)
Chris Giese     <geezer@execpc.com>     http://my.execpc.com/~geezer
This code is public domain (no copyright).
You can do whatever you want with it.

March 22, 2012:
- replaced pci_t struct with 16-bit 'bdf' value
- removed "_config" from function names to shorten them
- pci_read_(blah) now returns value read instead of useless error code
- pci_write_(blah) no longer returns useless error code

August 27, 2011:
- changed device validation code
- stylistic changes to pci_iterate()

July 23, 2009: modified to display device API

April 25, 2007: initial release
----------------------------------------------------------------------------*/
#include <string.h>
#include <stdio.h>

#if defined(__DJGPP__)
#include <dos.h> /* inport[b|w|l](), outport[b|w|l]() */

#elif defined(__WATCOMC__)
#include <conio.h>

#define inportb(P)      inp(P)
#define inportw(P)      inpw(P)
#define outportb(P,V)   outp(P,V)
#define outportw(P,V)   outpw(P,V)
#if defined(__386__)
#define inportl(P)      inpd(P)
#define outportl(P,V)   outpd(P,V)
/* else [in|out]portl() defined below */
#endif

#elif defined(__TURBOC__)
#include <dos.h> /* inport[b](), outport[b]() */

#define inportw(P)      inport(P)
#define outportw(P,V)   outport(P,V)
/* [in|out]portl() defined below */

#else
#include <conio.h>
#endif

#define PCI_ADDR_REG            0xCF8
#define PCI_DATA_REG            0xCFC
/*****************************************************************************
These functions are _code_, but are stored in the _data_ segment.
Declare them 'far' and end them with 'retf' instead of 'ret'.

For 16-bit Watcom C, use 'cdecl' to force usage of normal, stack
calling convention instead of Watcom register calling convention.
*****************************************************************************/
#if defined(__TURBOC__) || (defined(__WATCOMC__)&&!defined(__386__))
static const unsigned char g_inportl[] =
{                               /* BITS 16 */
        0x55,                   /* push bp */
        0x8B, 0xEC,             /*  mov bp,sp */
        0x8B, 0x56, 0x06,       /*  mov dx,[bp + 6] */
        0x66, 0xED,             /*  in eax,dx */
        0x8B, 0xD0,             /*  mov dx,ax */
        0x66, 0xC1, 0xE8, 0x10, /*  shr eax,16 */
        0x92,                   /*  xchg dx,ax */
        0x5D,                   /* pop bp */
        0xCB                    /* retf */
};

static unsigned long far cdecl (*inportl)(unsigned port) =
        (unsigned long far cdecl (*)(unsigned))g_inportl;
/*****************************************************************************
*****************************************************************************/
static const unsigned char g_outportl[] =
{                               /* BITS 16 */
        0x55,                   /* push bp */
        0x8B, 0xEC,             /*  mov bp,sp */
        0x8B, 0x56, 0x06,       /*  mov dx,[bp + 6] */
        0x66, 0x8B, 0x46, 0x08, /*  mov eax,[bp + 8] */
        0x66, 0xEF,             /*  out dx,eax */
        0x5D,                   /* pop bp */
        0xCB                    /* retf */
};

static void far cdecl (*outportl)(unsigned port, unsigned long val) =
        (void far cdecl (*)(unsigned, unsigned long))g_outportl;
#endif
/*****************************************************************************
*****************************************************************************/
static int pci_detect(void)
{
        printf("PCI controller...");
/* poke 32-bit I/O register at 0xCF8 to see if
there's a PCI controller there */
        outportl(PCI_ADDR_REG, 0x80000000L);
        if(inportl(PCI_ADDR_REG) != 0x80000000L)
        {
                printf("not found\n");
                return -1;
        }
        printf("found\n");
        return 0;
}
/*****************************************************************************
'bdf' = bus (b15-b8), device (b7-b3), and function (b2-b0)

Bitfields for PCI configuration address port:
Bit(s)  Description     (Table P0944)
 1-0    reserved (00)
 7-2    configuration register number (see #00878)
 10-8   function
 15-11  device number
 23-16  bus number
 30-24  reserved (0)
 31     enable configuration space mapping
*****************************************************************************/
static unsigned pci_read_byte(unsigned bdf, unsigned reg)
{
/* Turbo C shift bug */
#if 0
        outportl(PCI_ADDR_REG, 0x80000000L /* "enable configuration space mapping" */
                | ((unsigned long)bdf << 8) | (reg & ~3));
#else
        unsigned long i = bdf;

        i <<= 8;
        i |= 0x80000000L; /* "enable configuration space mapping" */
        i |= (reg & ~3);
        outportl(PCI_ADDR_REG, i);
#endif
        return inportb(PCI_DATA_REG + (reg & 3));
}
/*****************************************************************************
note that b0 of 'reg' is ignored
*****************************************************************************/
static unsigned pci_read_word(unsigned bdf, unsigned reg)
{
/*      outportl(PCI_ADDR_REG, 0x80000000L
                | ((unsigned long)bdf << 8) | (reg & ~3)); */
        unsigned long i = bdf;

        i <<= 8;
        i |= 0x80000000L;
        i |= (reg & ~3);
        outportl(PCI_ADDR_REG, i);
        return inportw(PCI_DATA_REG + (reg & 2));
}
/*****************************************************************************
note that b0 and b1 of 'reg' are ignored
*****************************************************************************/
static unsigned long pci_read_dword(unsigned bdf, unsigned reg)
{
/*      outportl(PCI_ADDR_REG, 0x80000000L
                | ((unsigned long)bdf << 8) | (reg & ~3)); */
        unsigned long i = bdf;

        i <<= 8;
        i |= 0x80000000L;
        i |= (reg & ~3);
        outportl(PCI_ADDR_REG, i);
        return inportl(PCI_DATA_REG);
}
/*****************************************************************************
*****************************************************************************/
static void pci_write_byte(unsigned bdf, unsigned reg, unsigned val)
{
/*      outportl(PCI_ADDR_REG, 0x80000000L
                | ((unsigned long)bdf << 8) | (reg & ~3)); */
        unsigned long i = bdf;

        i <<= 8;
        i |= 0x80000000L;
        i |= (reg & ~3);
        outportl(PCI_ADDR_REG, i);
// xxx - is this right?
        outportb(PCI_DATA_REG + (reg & 3), val);
}
/*****************************************************************************
note that b0 of 'reg' is ignored
*****************************************************************************/
static void pci_write_word(unsigned bdf, unsigned reg, unsigned val)
{
/*      outportl(PCI_ADDR_REG, 0x80000000L
                | ((unsigned long)bdf << 8) | (reg & ~3)); */
        unsigned long i = bdf;

        i <<= 8;
        i |= 0x80000000L;
        i |= (reg & ~3);
        outportl(PCI_ADDR_REG, i);
        outportw(PCI_DATA_REG + (reg & 2), val);
}
/*****************************************************************************
note that b0 and b1 of 'reg' are ignored
*****************************************************************************/
static void pci_write_dword(unsigned bdf, unsigned reg, unsigned long val)
{
/*      outportl(PCI_ADDR_REG, 0x80000000L
                | ((unsigned long)bdf << 8) | (reg & ~3)); */
        unsigned long i = bdf;

        i <<= 8;
        i |= 0x80000000L;
        i |= (reg & ~3);
        outportl(PCI_ADDR_REG, i);
        outportl(PCI_DATA_REG, val);
}
/*****************************************************************************
'bdf' = bus (b15-b8), device (b7-b3), and function (b2-b0)
*****************************************************************************/
#define FN(BDF)         (((BDF) >> 0) & 0x03)
#define DEV(BDF)        (((BDF) >> 3) & 0x1F)
#define BUS(BDF)        (((BDF) >> 8) & 0xFF)


static int pci_iterate(unsigned *bdf_ptr)
{
        unsigned char hdr_type, mf = 0;
        unsigned bdf = *bdf_ptr, next;

/* if first function of device, check if multi-function device
(otherwise fn==0 is the _only_ function of this device) */
        if(FN(bdf) == 0)
        {
/* 0x0E=PCI_HEADER_TYPE (byte) */
                hdr_type = pci_read_byte(bdf, 0x0E);
/* error in spec? I think they got this bit inverted...
"Bit 7 in this register is used to identify a multi-function device.
If the bit is 0, then the device is single function. If the bit is 1,
then the device has multiple functions.
                mf = hdr_type & 0x80; */
                mf = ((hdr_type & 0x80) == 0);
        }
/* advance to next device if current device is NOT multi-function... */
        if(mf)
                next = bdf + 0x08;
/* ...else advance to next function in current device
(and next device if current function = 7) */
        else
                next = bdf + 1;
/* (...and advance to next bus if current device = 31) */

        //if(BUS(next) != BUS(bdf) && BUS(next) > g_last_pci_bus)

        //if(next == 0) return 1; /* done */
        if(next > 999) return 1;  // GB 2014, make it stop

        *bdf_ptr = next;

        return 0;
}
/*****************************************************************************
"class" is a C++ reserved word
*****************************************************************************/
static void display_device_class(unsigned _class, unsigned subclass)
{
        static const char *name1[] =
        {
                "SCSI", "IDE", "floppy", "IPI", "RAID"
        };
        static const char *name2[] =
        {
                "Ethernet", "Token Ring", "FDDI", "ATM"
        };
        static const char *name3[] =
        {
                "VGA", "SuperVGA", "XGA"
        };
        static const char *name4[] =
        {
                "video", "audio"
        };
        static const char *name5[] =
        {
                "RAM", "Flash"
        };
        static const char *name6[] =
        {
                "CPU", "ISA", "EISA", "MicroChannel", "PCI", "PCMCIA",
                "NuBus", "CardBus"
        };
        static const char *name7[] =
        {
                "PC serial", "PC parallel"
        };
        static const char *name8[] =
        {
                "8259 PIC", "8237 DMAC", "8254 PTC", "MC146818 RTC"
        };
        static const char *name9[] =
        {
                "keyboard", "digitizer/pen", "mouse"
        };
        static const char *name10[] =
        {
                "generic"
        };
        static const char *name11[] =
        {
                "386", "486", "Pentium", "Pentium Pro (P6)"
/* 0x10=DEC Alpha; 0x40=coprocessor */
        };
        static const char *name12[] =
        {
                "Firewire (IEEE 1394)", "ACCESS.bus", "SSA", "USB",
                "Fiber Channel"
        };
/**/

        switch(_class)
        {
        case 1:
                printf("disk controller");
                if(subclass >= sizeof(name1) / sizeof(name1[0]))
ERR:            {
                        printf(":unknown (subclass=%u)", subclass);
                        return;
                }
                printf(":%s", name1[subclass]);
                break;
        case 2:
                printf("network controller");
                if(subclass >= sizeof(name2) / sizeof(name2[0]))
                        goto ERR;
                printf(":%s", name2[subclass]);
                break;
        case 3:
                printf("display controller");
                if(subclass >= sizeof(name3) / sizeof(name3[0]))
                        goto ERR;
                printf(":%s", name3[subclass]);
                break;
        case 4:
                printf("multimedia controller");
                if(subclass >= sizeof(name4) / sizeof(name4[0]))
                        goto ERR;
                printf(":%s", name4[subclass]);
                break;
        case 5:
                printf("memory");
                if(subclass >= sizeof(name5) / sizeof(name5[0]))
                        goto ERR;
                printf(":%s", name5[subclass]);
                break;
        case 6:
                printf("bridge");
                if(subclass >= sizeof(name6) / sizeof(name6[0]))
                        goto ERR;
                printf(":%s", name6[subclass]);
                break;
        case 7:
                printf("communications device");
                if(subclass >= sizeof(name7) / sizeof(name7[0]))
                        goto ERR;
                printf(":%s", name7[subclass]);
                break;
        case 8:
                printf("system device");
                if(subclass >= sizeof(name8) / sizeof(name8[0]))
                        goto ERR;
                printf(":%s", name8[subclass]);
                break;
        case 9:
                printf("HID");
                if(subclass >= sizeof(name9) / sizeof(name9[0]))
                        goto ERR;
                printf(":%s", name9[subclass]);
                break;
        case 10:
                printf("dock");
                if(subclass >= sizeof(name10) / sizeof(name10[0]))
                        goto ERR;
                printf(":%s", name10[subclass]);
                break;
        case 11:
                printf("CPU");
                if(subclass >= sizeof(name11) / sizeof(name11[0]))
                        goto ERR;
                printf(":%s", name11[subclass]);
                break;
        case 12:
                printf("serial bus controller");
                if(subclass >= sizeof(name12) / sizeof(name12[0]))
                        goto ERR;
                printf(":%s", name12[subclass]);
                break;
        default:
                printf("unknown (class=%u)", _class);
                break;
        }
}

//------------------------------------------------------------------------------

int header;

int pci_scan_and_set(int target, unsigned *outbdf)
{
    unsigned long i;
    unsigned bdf;
    
    unsigned int val=0;                // Variable

    /* check for PCI */
    /* display numeric ID of all PCI devices detected */
    bdf = 0;

        *outbdf = NULL;

    do
    {
                unsigned vendor, device, _class, subclass, api;

        /* 00=PCI_VENDOR_ID (word), 02=PCI_DEVICE_ID (word) */
        i = pci_read_dword(bdf, 0x00);
        //printf("%d -> %lX\n", bdf, i);
        vendor = (unsigned)(i & 0xFFFF);
        device = (unsigned)(i >> 16);
        /* anything there?
        if(i != 0xFFFFFFFFL)
        if(vendor != 0 && vendor != 0xFFFF)
        "Since 0FFFFh is an invalid Vendor ID, it is adequate for the host bus to
         PCI bridge to return a value of all 1's on read accesses to Configratuion
        Space registers of non-existent devices." */
        if(vendor != 0xFFFF)
        {
            if (vendor==0x10B9 && device==target)
            {
               if (header==0)
               printf("Bus Device Function VendorID DeviceID Class:Subclass:API\n"
                      "--- ------ -------- -------- -------- ------------------\n");
               header=1;

               printf("%3u %6u %8u %8X %-8X ", BUS(bdf),
                      DEV(bdf), FN(bdf), vendor, device);
               /* 09=PCI_PROGRAMMING_INTERFACE, 0A=PCI_SUBCLASS, 0B=PCI_CLASS */
               i = pci_read_dword(bdf, 0x08);
               _class   = (unsigned)((i >> 24) & 0xFF);
               subclass = (unsigned)((i >> 16) & 0xFF);
               api      = (unsigned)((i >>  8) & 0xFF);

               display_device_class(_class, subclass);
               printf(":%02X %08X\n", api, bdf);
   
                           *outbdf = bdf;
               return 1; // done
/*
               if (vendor==0x10B9 && device==0x7101) 
               {           
                   printf("                                      ");             
                   val=pci_read_byte(bdf, 0x05B);
                   if(val & 64) // bit 6 is set, means disable
                   {
                       printf("-> Enabling Throttle\n");
                       val = val -64;
                       pci_write_byte(bdf, 0x05B, val);
                   }
                   else  
                       printf("-> ALi m7101 throttle already enabled.\n");
                }

                if (vendor==0x10B9 && device==0x7101) 
                {           
                    printf("                                      ");             
                    val=pci_read_byte(bdf, 0x07B);
                    if (!(val & 4)) // bit 2 is not set, needs to be set 
                    {
                        printf("-> Enabling STPCLK\n");
                        val = val +4;
                        pci_write_byte(bdf, 0x07B, val);
                    }
                    else  
                        printf("-> ALi m7101 STPCLK already enabled.\n");
                }
*/

                #if 0
                /* I need to run this code on other systems, e.g. with AGP bus... */
                {
                                        unsigned char blah;
                    /* 0x0E=PCI_HEADER_TYPE (byte) */
                    blah = pci_read_byte(bdf, 0x0E);
                    if((blah & 0x7F) == 1)
                    {
                        i = pci_read_dword(bdf, 0x18);
                        printf("    primary bus=%lu, secondary bus=%lu, subordinate bus=%lu\n",
                        (i >> 0) & 0xFF, (i >> 8) & 0xFF, (i >> 16) & 0xFF);
                    }
                }
                #endif
/*
                if(_class == 12 && subclass == 3)
                {
                    unsigned pi;

                    printf("USB controller detected; "
                           "programming interface=");
                    pi = (unsigned)((i >> 8) & 0xFF);
                    if(pi == 0)
                        printf("UHCI\n");
                    else if(pi == 0x10)
                        printf("OHCI\n");
                    else if(pi == 0x20)
                        printf("EHCI\n");
                    else
                        printf("unknown (0x%02X)\n", pi);
                }
*/
            }
        }
        //getchar();
    } 
    while(!pci_iterate(&bdf));
    return 0;
}

// Find an empty port address for PMU IO
unsigned find_base_addr()
{
        unsigned addr, i, good;
        for(addr = 0x9000; addr < 0xF000; addr += 0x100)
        {
                good = 1;
                for(i = 0;i < 256;i += 2)
                {
                        unsigned result;
                        result = inportw(addr+i);
                        if(result != 0xFFFFu)
                        {
                                good = 0;
                                break;
                        }
                }

                if(good)
                        return addr;
        }
        return 0;
}

// Sets the PMU IO base address
void program_base_addr(unsigned regbdf, unsigned addr)
{
        unsigned char val;

        // Disable IO
        val = pci_read_byte(regbdf, 0x4);
        val &= ~1;
        pci_write_byte(regbdf, 0x4, val);

        // Porgram address
        addr = (addr&0xFFC0u)|1;
        pci_write_word(regbdf, 0x10, addr);

        // Enable IO
        val = pci_read_byte(regbdf, 0x4);
        val |= 1;
        pci_write_byte(regbdf, 0x4, val);
}

// GB 2014
// Checks whether PMU and SMB are enabled and turns them on in case they are not. 
// It's done by clearing Bit 2 in M1533 config space 5Fh.
int unlock_pmu(unsigned pmubdf)
{
        unsigned char val;

        val = pci_read_byte(pmubdf, 0x5F);
        if (val & 0x4) 
        {
                val = val & 0xFB;
                pci_write_byte(pmubdf, 0x5F, val);
                val = pci_read_byte(pmubdf, 0x5F);
        }

        return (val & 0x4) == 0x4;
}

#define ALI15X3_LOCK 0x06  // ALI15X3 address lock bits
int do_throttle(unsigned pmubdf, unsigned regbdf)
{
        int amount;
        unsigned val, baseaddr;

        baseaddr = find_base_addr();

        puts("\nSpeed options:");
        puts("0. top speed              (No throttle)");
        puts("1. not really slow at all (12.5% throttled)");
        puts("2. sorta kinda slow       (25% throttled)");
        puts("3. gettin' there          (37.5% throttled)");
        puts("4. half slow              (50% throttled)");
        puts("5. pretty slow            (62.5% throttled)");
        puts("6. considerably slow      (75% throttled)");
        puts("7. whoa! Is it 1993?      (87.5% throttled)");
        puts("ESC. get me outta here. No changes.\n");

        amount = getch();
        if(amount < '0' || amount > '7')
                return 0;

        amount -= '0';

        // Recheck in case it was relocked
        unlock_pmu(pmubdf);

        val = pci_read_byte(regbdf, 0x05B);
        if(val & ALI15X3_LOCK) // bit 1 and 2 are set (value 6)
        {
                printf("Unlocking ALi m7101 address registers\n");
                val &= ~ALI15X3_LOCK;
                pci_write_byte(regbdf, 0x05B, val);
        }
        else
                printf("ALi m7101 registers already unlocked.\n");

        program_base_addr(regbdf, baseaddr);
        val = inportw(baseaddr+0x10);

                val &= ~0x21E;
        outportw(baseaddr+0x10, val);

        if(amount > 0)
        {
                val |= ((8-amount)<<1)|0x210; // BIT4 = THRO_EN, BIT9 = CLN_EN
                outportw(baseaddr+0x10, val);
                                printf("Throttle enabled: %04X\n", val);
        }
                else
                {
                        printf("Throttle disabled: %04X\n", val);
                }
                return 0;
}

/*****************************************************************************
*****************************************************************************/
int main(void)
{
   unsigned pmubdf, regbdf;
   int returncode=0;
   header=0;
   printf("ALi Aladdin V ACPI Throttle - G. Broers, B. Obrzut, thanks to Chris Giese.\n");
   if (pci_detect()) return 1;
   returncode=pci_scan_and_set(0x1533, &pmubdf);
   if (returncode)
   {
           unlock_pmu(pmubdf);
           returncode=pci_scan_and_set(0x7101, &regbdf);
   }
   if (returncode==0)
   {
           printf("No m7101 PMU Detected.");
           return 1;
   }

   returncode = do_throttle(pmubdf, regbdf);
   return returncode; // returns 1 if found 
}
